@@ -52,6 +52,14 @@
 #define PTRACE_INTERRUPT 0x4207
 #endif
 
+/*
+ * The ustar format only has 11 octal characters available for
+ * specifying sizes and offsets. So the maximum value is:
+ *
+ *       077777777777 => 8589934591
+ */
+#define USTAR_MAXVAL 8589934591
+
 static struct dump_info *global_di;
 static long PAGESZ;
 
@@ -761,7 +769,7 @@ static int init_log(struct dump_info *di)
 typedef int elf_parse_cb(struct dump_info *di, Elf *elf, GElf_Phdr *phdr);
 
 static int do_elf_ph_parse(struct dump_info *di, GElf_Phdr *type,
-			   elf_parse_cb *callback)
+			   elf_parse_cb *callback, size_t *phnum_found)
 {
 	GElf_Ehdr ehdr_mem;
 	GElf_Ehdr *ehdr;
@@ -770,6 +778,9 @@ static int do_elf_ph_parse(struct dump_info *di, GElf_Phdr *type,
 	size_t phnum;
 	size_t cnt;
 
+	if (phnum_found)
+		*phnum_found = 0;
+
 	/* start from beginning of core */
 	if (lseek64(di->elf_fd, 0, SEEK_SET) == -1) {
 		info("lseek failed: %s", strerror(errno));
@@ -809,6 +820,9 @@ static int do_elf_ph_parse(struct dump_info *di, GElf_Phdr *type,
 		goto out;
 	}
 
+	if (phnum_found)
+		*phnum_found = phnum;
+
 	for (cnt = 0; cnt < phnum; cnt++) {
 		GElf_Phdr phdr_mem;
 		GElf_Phdr *phdr;
@@ -891,7 +905,7 @@ static int vma_cb(struct dump_info *di, Elf *elf, GElf_Phdr *phdr)
 /*
  * Tries to parse the found ELF headers and reads all vmas from it.
  */
-static int parse_vma_info(struct dump_info *di)
+static int parse_vma_info(struct dump_info *di, size_t *phnum_found)
 {
 	unsigned long min_off = ULONG_MAX;
 	unsigned long max_len = 0;
@@ -911,7 +925,7 @@ static int parse_vma_info(struct dump_info *di)
 	memset(&type, 0, sizeof(type));
 	type.p_type = PT_LOAD;
 	type.p_flags = PF_R;
-	if (do_elf_ph_parse(di, &type, vma_cb) != 0)
+	if (do_elf_ph_parse(di, &type, vma_cb, phnum_found) != 0)
 		return -1;
 
 	for (v = di->vma; v; v = v->next) {
@@ -966,6 +980,14 @@ static int copy_data(int src, int dest, int dest2, size_t len, char *pagebuf)
 		if (ret < 0) {
 			info("read core failed at 0x%lx",
 			     lseek64(src, 0, SEEK_CUR));
+
+			/* skip this chunk */
+			lseek64(src, chunk, SEEK_CUR);
+			memset(pagebuf, 0, chunk);
+
+		} else if (ret == 0) {
+			info("read core eof-failed at 0x%lx",
+			     lseek64(src, 0, SEEK_CUR));
 			return -1;
 		}
 
@@ -1106,10 +1128,15 @@ static off64_t block_roundup(off64_t b)
 
 static int dump_zero(int fd, off64_t count)
 {
+	static char zero_block[4096];
+	size_t sz = sizeof(zero_block);
+
 	while (count) {
-		if (write_file_fd(fd, "", 1) < 0)
+		if (count < sizeof(zero_block))
+			sz = count;
+		if (write_file_fd(fd, zero_block, sz) < 0)
 			return -1;
-		count--;
+		count -= sz;
 	}
 
 	return 0;
@@ -1494,6 +1521,15 @@ int add_core_data(struct dump_info *di, off64_t dest_offset, size_t len,
 
 	end = start + len;
 
+	if (di->cfg->prog_config.core_in_tar &&
+	    di->cfg->prog_config.core_compressor &&
+	    (start > USTAR_MAXVAL || end > USTAR_MAXVAL)) {
+		info("core data too large for ustar format "
+		     "(0x%" PRIx64 "-0x%" PRIx64 "), dropping",
+		     start, end);
+		return EFBIG;
+	}
+
 	for (cur = di->core_file; cur && !done; cur = cur->next) {
 		if (end < cur->start) {
 			/* insert new block */
@@ -1608,14 +1644,36 @@ int add_core_data(struct dump_info *di, off64_t dest_offset, size_t len,
 	return 0;
 }
 
+/*
+ * In case the core file is packed into a tar, make sure the core file
+ * size does not exceed the value limits of the ustar format.
+ */
+static void check_core_size(struct dump_info *di)
+{
+	if (!di->cfg->prog_config.core_in_tar)
+		return;
+	if (!di->cfg->prog_config.core_compressor)
+		return;
+	if (di->core_file_size <= USTAR_MAXVAL)
+		return;
+
+	info("core is too large for ustar format (%" PRIu64 " bytes), "
+	     "truncating to %" PRIu64 " bytes",
+	     di->core_file_size, USTAR_MAXVAL);
+
+	di->core_file_size = USTAR_MAXVAL;
+}
+
 /*
  * Reads the ELF header from the large core file.
  * This header is dumped to the core.
  */
 static int init_src_core(struct dump_info *di, int src)
 {
+	size_t last_phnum = 0;
 	int tries = 0;
 	int ret = -1;
+	size_t phnum;
 	size_t len;
 	char *buf;
 	long pos;
@@ -1642,7 +1700,7 @@ again:
 		goto out;
 
 	/* try to elf-parse the core to read vma info */
-	ret = parse_vma_info(di);
+	ret = parse_vma_info(di, &phnum);
 
 	/* restore our position */
 	if (lseek64(di->elf_fd, pos, SEEK_SET) == -1)
@@ -1653,9 +1711,17 @@ again:
 
 		tries++;
 
-		/* maybe try again */
-		if (tries < 10)
+		if (phnum > last_phnum) {
+			/* new headers found, keep trying */
+			last_phnum = phnum;
 			goto again;
+		} else if (tries < 10) {
+			/*
+			 * even if no new headers are found,
+			 * retry at least 10 times
+			 */
+			goto again;
+		}
 
 		goto out;
 	}
@@ -1673,6 +1739,7 @@ again:
 
 	/* make the core big enough to fit all vma areas */
 	di->core_file_size = di->vma_end;
+	check_core_size(di);
 
 	/* add empty core data to mark the size of the core file */
 	add_core_data(di, di->core_file_size, 0, di->elf_fd, 0);
@@ -2018,8 +2085,8 @@ static int dump_vma(struct dump_info *di, unsigned long start, size_t len,
 		/* only dump what is actually in VMA */
 		if (dump_start < tmp->start)
 			dump_start = tmp->start;
-		if (dump_end > tmp->mem_end)
-			dump_end = tmp->mem_end;
+		if (dump_end > tmp->file_end)
+			dump_end = tmp->file_end;
 
 		/* make sure we have something to dump */
 		if (dump_start < dump_end) {
@@ -2106,7 +2173,7 @@ static int dump_stacks(struct dump_info *di)
 		/* find and set the first task */
 		memset(&type, 0, sizeof(type));
 		type.p_type = PT_NOTE;
-		do_elf_ph_parse(di, &type, note_cb);
+		do_elf_ph_parse(di, &type, note_cb, NULL);
 	}
 
 	if (di->first_pid)
@@ -2269,13 +2336,27 @@ out_err:
 #undef MAPS_LINE_MAXSIZE
 }
 
-static int read_remote(struct dump_info *di, unsigned long addr, void *dst,
-		       ssize_t len)
+static int do_exact_pread(int fd, void *buf, size_t nbytes, off64_t offset)
 {
 	int ret;
 
-	ret = pread64(di->mem_fd, dst, len, addr);
-	if (ret != len) {
+	while (nbytes != 0) {
+		ret = pread64(fd, buf, nbytes, offset);
+		if (ret == 0 || ret == -1)
+			return -1;
+
+		buf += ret;
+		offset += ret;
+		nbytes -= ret;
+	}
+
+	return 0;
+}
+
+static int read_remote(struct dump_info *di, unsigned long addr, void *dst,
+		       ssize_t len)
+{
+	if (do_exact_pread(di->mem_fd, dst, len, addr) != 0) {
 		info("read_remote failed: len=%d, addr=0x%lx, "
 		     "dest=0x%x, errno=\"%s\"",
 		     len, addr, dst, strerror(errno));
@@ -2305,8 +2386,7 @@ static int alloc_remote_string(struct dump_info *di, unsigned long addr,
 		return ENOMEM;
 
 	for (i = 1; i < REMOTE_STRING_MAX; i++) {
-		ret = pread64(di->mem_fd, ptr, i, addr);
-		if (ret != i) {
+		if (do_exact_pread(di->mem_fd, ptr, i, addr) != 0) {
 			ret = errno;
 			info("read_remote failed: addr %#lx: %s", addr,
 			     strerror(errno));
@@ -3000,8 +3080,10 @@ static void get_pthread_list_fallback(struct dump_info *di)
 	unsigned long addr;
 
 	/* try to determine the size of "struct pthread" */
-	if (sym_address(di, "_thread_db_sizeof_pthread", &addr) == 0)
-		read_remote(di, addr, &pthreadsz, sizeof(pthreadsz));
+	if (sym_address(di, "_thread_db_sizeof_pthread", &addr) == 0) {
+		if (read_remote(di, addr, &pthreadsz, sizeof(pthreadsz)) != 0)
+			pthreadsz = 0;
+	}
 	if (pthreadsz == 0) {
 		pthreadsz = PAGESZ;
 		info("guessing sizeof(struct pthread): %u bytes", pthreadsz);
@@ -3168,7 +3250,8 @@ static int init_from_auxv(struct dump_info *di, ElfW(auxv_t) *auxv,
 		/* val32 = (ElfW(Phdr))phdr_addr[i].p_type */
 		addr = phdr_addr + (sizeof(ElfW(Phdr)) * i) +
 		       offsetof(ElfW(Phdr), p_type);
-		read_remote(di, addr, &val32, sizeof(val32));
+		if (read_remote(di, addr, &val32, sizeof(val32)) != 0)
+			break;
 
 		if (val32 == PT_NULL) {
 			break;
@@ -3176,7 +3259,10 @@ static int init_from_auxv(struct dump_info *di, ElfW(auxv_t) *auxv,
 		} else if (val32 == PT_PHDR) {
 			addr = phdr_addr + (sizeof(ElfW(Phdr)) * i) +
 			       offsetof(ElfW(Phdr), p_vaddr);
-			read_remote(di, addr, &relocation, sizeof(relocation));
+			if (read_remote(di, addr, &relocation,
+					sizeof(relocation)) != 0) {
+				break;
+			}
 			found |= 0x1;
 
 			relocation = phdr_addr - relocation;
@@ -3185,7 +3271,10 @@ static int init_from_auxv(struct dump_info *di, ElfW(auxv_t) *auxv,
 			/* dyn_addr = (ElfW(Phdr))phdr_addr[i].p_vaddr */
 			addr = phdr_addr + (sizeof(ElfW(Phdr)) * i) +
 			       offsetof(ElfW(Phdr), p_vaddr);
-			read_remote(di, addr, &dyn_addr, sizeof(dyn_addr));
+			if (read_remote(di, addr, &dyn_addr,
+					sizeof(dyn_addr)) != 0) {
+				break;
+			}
 			found |= 0x2;
 		}
 	}
@@ -3212,7 +3301,8 @@ static int init_from_auxv(struct dump_info *di, ElfW(auxv_t) *auxv,
 		/* val32 = (ElfW(Dyn))dyn_addr[i].d_tag */
 		addr = dyn_addr + (sizeof(ElfW(Dyn)) * i)
 		       + offsetof(ElfW(Dyn), d_tag);
-		read_remote(di, addr, &val32, sizeof(val32));
+		if (read_remote(di, addr, &val32, sizeof(val32)) != 0)
+			break;
 
 		if (val32 == DT_NULL) {
 			break;
@@ -3221,7 +3311,10 @@ static int init_from_auxv(struct dump_info *di, ElfW(auxv_t) *auxv,
 			/* debug_ptr = (ElfW(Dyn))dyn_addr[i].d_un.d_ptr */
 			addr = dyn_addr + (sizeof(ElfW(Dyn)) * i) +
 			       offsetof(ElfW(Dyn), d_un.d_ptr);
-			read_remote(di, addr, debug_ptr, sizeof(*debug_ptr));
+			if (read_remote(di, addr, debug_ptr,
+					sizeof(*debug_ptr)) != 0) {
+				break;
+			}
 
 			/* found it! */
 			found |= 0x4;
@@ -3283,8 +3376,10 @@ static int get_so_list(struct dump_info *di)
 		dump_vma(di, ptr, sizeof(struct r_debug), 0, "auxv r_debug");
 
 	/* get pointer to first link_map */
-	read_remote(di, ptr + offsetof(struct r_debug, r_map), &ptr,
-		    sizeof(ptr));
+	if (read_remote(di, ptr + offsetof(struct r_debug, r_map), &ptr,
+			sizeof(ptr)) != 0) {
+		return -1;
+	}
 
 	while (ptr) {
 		unsigned long addr = 0;
@@ -3297,8 +3392,10 @@ static int get_so_list(struct dump_info *di)
 		}
 
 		/* get pointer to link_map name */
-		read_remote(di, ptr + offsetof(struct link_map, l_name),
-			    &addr, sizeof(addr));
+		if (read_remote(di, ptr + offsetof(struct link_map, l_name),
+				&addr, sizeof(addr)) != 0) {
+			return -1;
+		}
 
 		if (alloc_remote_string(di, addr, &l_name) == 0) {
 			/* dump link_map name */
@@ -3310,20 +3407,22 @@ static int get_so_list(struct dump_info *di)
 			/* store so data since we are here */
 			if (l_name[0] != 0) {
 				/* get pointer to base address */
-				read_remote(di,
-					ptr + offsetof(struct link_map,
-						       l_addr),
-					&addr, sizeof(addr));
-
-				store_sym_data(di, l_name, addr);
+				if (read_remote(di,
+						ptr + offsetof(struct link_map,
+							       l_addr),
+						&addr, sizeof(addr)) == 0) {
+					store_sym_data(di, l_name, addr);
+				}
 			}
 
 			free(l_name);
 		}
 
 		/* get pointer to next link_map */
-		read_remote(di, ptr + offsetof(struct link_map, l_next),
-			    &ptr, sizeof(ptr));
+		if (read_remote(di, ptr + offsetof(struct link_map, l_next),
+				&ptr, sizeof(ptr)) != 0) {
+			return -1;
+		}
 	}
 
 	return 0;
@@ -3401,6 +3500,7 @@ static int add_dumplist_section(struct dump_info *di)
 	}
 
 	di->core_file_size = core_size;
+	check_core_size(di);
 
 	add_core_data(di, dump_offset, core_size - dump_offset,
 		      di->elf_fd, dump_offset);
